# flake8: noqa
from firedantic._async.helpers import truncate_collection as async_truncate_collection
from firedantic._async.model import (
    AsyncBareModel,
    AsyncBareSubCollection,
    AsyncBareSubModel,
    AsyncModel,
    AsyncSubCollection,
    AsyncSubModel,
)
from firedantic._sync.helpers import truncate_collection
from firedantic._sync.model import (
    BareModel,
    BareSubCollection,
    BareSubModel,
    Model,
    SubCollection,
    SubModel,
)
from firedantic.configurations import CONFIGURATIONS, configure
from firedantic.exceptions import *



from google.cloud.firestore_v1 import AsyncCollectionReference


async def truncate_collection(
    col_ref: AsyncCollectionReference, batch_size: int = 128
) -> int:
    """Removes all documents inside a collection.

    :param batch_size: Batch size for listing documents.
    :return: Number of removed documents.
    """
    count = 0

    while True:
        deleted = 0
        async for doc in col_ref.limit(batch_size).stream():  # type: ignore
            await doc.reference.delete()
            deleted += 1

        count += deleted
        if deleted < batch_size:
            return count



class ModelError(Exception):
    """Generic model error class."""

    pass


class InvalidDocumentID(ModelError):
    pass


class ModelNotFoundError(ModelError):
    pass


class CollectionNotDefined(ModelError):
    pass



from typing import Any, Dict, Union

from google.cloud.firestore_v1 import AsyncClient, Client

CONFIGURATIONS: Dict[str, Any] = {}


def configure(db: Union[Client, AsyncClient], prefix: str = "") -> None:
    """Configures the prefix and DB.

    :param db: The firestore client instance.
    :param prefix: The prefix to use for collection names.
    """
    global CONFIGURATIONS

    CONFIGURATIONS["db"] = db
    CONFIGURATIONS["prefix"] = prefix



from abc import ABC
from logging import getLogger
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import pydantic
from google.cloud.firestore_v1 import (
    CollectionReference,
    DocumentReference,
    DocumentSnapshot,
)
from google.cloud.firestore_v1.base_query import BaseQuery

import firedantic.operators as op
from firedantic import truncate_collection
from firedantic.configurations import CONFIGURATIONS
from firedantic.exceptions import (
    CollectionNotDefined,
    InvalidDocumentID,
    ModelNotFoundError,
)

TBareModel = TypeVar("TBareModel", bound="BareModel")
TBareSubModel = TypeVar("TBareSubModel", bound="BareSubModel")
logger = getLogger("firedantic")

# https://firebase.google.com/docs/firestore/query-data/queries#query_operators
FIND_TYPES = {
    op.LT,
    op.LTE,
    op.EQ,
    op.NE,
    op.GT,
    op.GTE,
    op.ARRAY_CONTAINS,
    op.ARRAY_CONTAINS_ANY,
    op.IN,
    op.NOT_IN,
}


def _get_col_ref(cls, name) -> CollectionReference:
    if name is None:
        raise CollectionNotDefined(f"Missing collection name for {cls.__name__}")

    collection: CollectionReference = CONFIGURATIONS["db"].collection(
        CONFIGURATIONS["prefix"] + name
    )
    return collection


class BareModel(pydantic.BaseModel, ABC):
    """Base model class.

    Implements basic functionality for Pydantic models, such as save, delete, find etc.
    """

    __collection__: Optional[str] = None
    __document_id__: str

    def save(self) -> None:
        """
        Saves this model in the database.

        :raise DocumentIDError: If the document ID is not valid.
        """
        data = self.dict(by_alias=True)
        if self.__document_id__ in data:
            del data[self.__document_id__]

        doc_ref = self._get_doc_ref()
        doc_ref.set(data)
        setattr(self, self.__document_id__, doc_ref.id)

    def delete(self) -> None:
        """
        Deletes this model from the database.

        :raise DocumentIDError: If the ID is not valid.
        """
        self._get_doc_ref().delete()

    def get_document_id(self):
        """
        Get the document ID for this model instance

        :raise DocumentIDError: If the ID is not valid.
        """
        doc_id = getattr(self, self.__document_id__, None)
        if doc_id is not None:
            self._validate_document_id(doc_id)
        return getattr(self, self.__document_id__, None)

    @classmethod
    def find(cls: Type[TBareModel], filter_: Optional[dict] = None) -> List[TBareModel]:
        """Returns a list of models from the database based on a filter.

        Example: `Company.find({"company_id": "1234567-8"})`.
        Example: `Product.find({"stock": {">=": 1}})`.

        :param filter_: The filter criteria.
        :return: List of found models.
        """
        if not filter_:
            filter_ = {}

        query: Union[BaseQuery, CollectionReference] = cls._get_col_ref()

        for key, value in filter_.items():
            query = cls._add_filter(query, key, value)

        def _cls(doc_id: str, data: Dict[str, Any]) -> TBareModel:
            if cls.__document_id__ in data:
                logger.warning(
                    "%s document ID %s contains conflicting %s in data with value %s",
                    cls.__name__,
                    doc_id,
                    cls.__document_id__,
                    data[cls.__document_id__],
                )
            data[cls.__document_id__] = doc_id
            model = cls(**data)
            setattr(model, cls.__document_id__, doc_id)
            return model

        return [
            _cls(doc_id, doc_dict)
            for doc_id, doc_dict in (
                (doc.id, doc.to_dict()) for doc in query.stream()  # type: ignore
            )
            if doc_dict is not None
        ]

    @classmethod
    def _add_filter(
        cls, query: Union[BaseQuery, CollectionReference], field: str, value: Any
    ) -> Union[BaseQuery, CollectionReference]:
        if type(value) is dict:
            for f_type in value:
                if f_type not in FIND_TYPES:
                    raise ValueError(
                        f"Unsupported filter type: {f_type}. Supported types are: {', '.join(FIND_TYPES)}"
                    )
                query: BaseQuery = query.where(field, f_type, value[f_type])  # type: ignore
            return query
        else:
            query: BaseQuery = query.where(field, "==", value)  # type: ignore
            return query

    @classmethod
    def find_one(cls: Type[TBareModel], filter_: Optional[dict] = None) -> TBareModel:
        """Returns one model from the DB based on a filter.

        :param filter_: The filter criteria.
        :return: The model instance.
        :raise ModelNotFoundError: If the entry is not found.
        """
        models = cls.find(filter_)
        try:
            return models[0]
        except IndexError:
            raise ModelNotFoundError(f"No '{cls.__name__}' found")

    @classmethod
    def get_by_doc_id(cls: Type[TBareModel], doc_id: str) -> TBareModel:
        """Returns a model based on the document ID.

        :param doc_id: The document ID of the entry.
        :return: The model.
        :raise ModelNotFoundError: Raised if no matching document is found.
        """

        try:
            cls._validate_document_id(doc_id)
        except InvalidDocumentID:
            # Getting a document with doc_id set to an empty string would raise a
            # google.api_core.exceptions.InvalidArgument exception and a doc_id
            # containing an uneven number of slashes would raise a
            # ValueError("A document must have an even number of path elements") and
            # could even load data from a sub collection instead of the desired one.
            raise ModelNotFoundError(
                f"No '{cls.__name__}' found with {cls.__document_id__} '{doc_id}'"
            )

        document: DocumentSnapshot = cls._get_col_ref().document(doc_id).get()  # type: ignore
        data = document.to_dict()
        if data is None:
            raise ModelNotFoundError(
                f"No '{cls.__name__}' found with {cls.__document_id__} '{doc_id}'"
            )
        data[cls.__document_id__] = doc_id
        model = cls(**data)
        setattr(model, cls.__document_id__, doc_id)
        return model

    @classmethod
    def truncate_collection(cls, batch_size: int = 128) -> int:
        """Removes all documents inside a collection.

        :param batch_size: Batch size for listing documents.
        :return: Number of removed documents.
        """
        return truncate_collection(
            col_ref=cls._get_col_ref(),
            batch_size=batch_size,
        )

    @classmethod
    def _get_col_ref(cls) -> CollectionReference:
        """Returns the collection reference."""
        return _get_col_ref(cls, cls.__collection__)

    def _get_doc_ref(self) -> DocumentReference:
        """
        Returns the document reference.

        :raise DocumentIDError: If the ID is not valid.
        """
        return self._get_col_ref().document(self.get_document_id())  # type: ignore

    @staticmethod
    def _validate_document_id(document_id: str):
        """
        Validates the Document ID is valid.

        Based on information from https://firebase.google.com/docs/firestore/quotas#limits

        :raise DocumentIDError: If the ID is not valid.
        """
        if len(document_id.encode("utf-8")) > 1500:
            raise InvalidDocumentID("Document ID must be no longer than 1,500 bytes")

        if "/" in document_id:
            raise InvalidDocumentID("Document ID cannot contain a forward slash (/)")

        if (
            document_id.startswith("__")
            and document_id.endswith("__")
            and len(document_id) >= 4
        ):
            raise InvalidDocumentID(
                "Document ID cannot match the regular expression __.*__"
            )

        if document_id in (".", ".."):
            raise InvalidDocumentID(
                "Document ID cannot solely consist of a single period (.) or double "
                "periods (..)"
            )

        if document_id == "":
            raise InvalidDocumentID("Document ID cannot be an empty string")


class Model(BareModel):
    __document_id__: str = "id"
    id: Optional[str] = None

    @classmethod
    def get_by_id(cls: Type[TBareModel], id_: str) -> TBareModel:
        return cls.get_by_doc_id(id_)


class BareSubCollection(ABC):
    __collection_tpl__: Optional[str] = None
    __document_id__: str

    @classmethod
    def model_for(cls, parent, model_class):
        parent_props = parent.dict(by_alias=True)

        name = model_class.__name__
        ic = type(name, (model_class,), {})
        ic.__collection_cls__ = cls
        ic.__collection__ = cls.__collection_tpl__.format(**parent_props)
        ic.__document_id__ = cls.__document_id__

        return ic


class BareSubModel(BareModel, ABC):
    __collection_cls__: "BareSubCollection"
    __collection__: Optional[str] = None
    __document_id__: str

    class Collection(BareSubCollection, ABC):
        pass

    @classmethod
    def _create(cls, **kwargs) -> TBareSubModel:
        return cls(  # type: ignore
            **kwargs,
        )

    @classmethod
    def _get_col_ref(cls) -> CollectionReference:
        """Returns the collection reference."""
        if cls.__collection__ is None or "{" in cls.__collection__:
            raise CollectionNotDefined(
                f"{cls.__name__} is not properly prepared. "
                f"You should use {cls.__name__}.model_for(parent)"
            )
        return _get_col_ref(cls.__collection_cls__, cls.__collection__)

    @classmethod
    def model_for(cls, parent):
        return cls.Collection.model_for(parent, cls)


class SubModel(BareSubModel):
    id: Optional[str] = None

    @classmethod
    def get_by_id(cls: Type[TBareModel], id_: str) -> TBareModel:
        """
        Get single item by document ID
        :raises ModelNotFoundError:
        """
        return cls.get_by_doc_id(id_)


class SubCollection(BareSubCollection, ABC):
    __document_id__ = "id"
    __model_cls__: Type[SubModel]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['firedantic',
 'firedantic._async',
 'firedantic._sync',
 'firedantic.tests',
 'firedantic.tests.tests_async',
 'firedantic.tests.tests_sync']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-firestore>=2.0.1,<3.0.0',
 'grpcio>=1.39.0,<2.0.0',
 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'firedantic',
    'version': '0.2.8',
    'description': 'Pydantic base models for Firestore',
    'long_description': '# Firedantic\n\n[![Build Status](https://travis-ci.com/digitalliving/firedantic.svg?branch=master)](https://travis-ci.com/digitalliving/firedantic)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyPI](https://img.shields.io/pypi/v/firedantic)](https://pypi.org/project/firedantic/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/firedantic)](https://pypi.org/project/firedantic/)\n[![License: BSD 3-Clause](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nDatabase models for Firestore using Pydantic base models.\n\n\n## Installation\n\nThe package is available on PyPi:\n\n```bash\npip install firedantic\n```\n\n\n## Usage\n\nIn your application you will need to configure the firestore db client and\noptionally the collection prefix, which by default is empty.\n\n```python\nfrom os import environ\nfrom unittest.mock import Mock\n\nimport google.auth.credentials\nfrom firedantic import configure\nfrom google.cloud.firestore import Client\n\n# Firestore emulator must be running if using locally.\nif environ.get("FIRESTORE_EMULATOR_HOST"):\n    client = Client(\n        project="firedantic-test",\n        credentials=Mock(spec=google.auth.credentials.Credentials)\n    )\nelse:\n    client = Client()\n\nconfigure(client, prefix="firedantic-test-")\n```\n\nOnce that is done, you can start defining your Pydantic models, e.g:\n\n```python\nfrom pydantic import BaseModel\n\nfrom firedantic import Model\n\nclass Owner(BaseModel):\n    """Dummy owner Pydantic model."""\n    first_name: str\n    last_name: str\n\n\nclass Company(Model):\n    """Dummy company Firedantic model."""\n    __collection__ = "companies"\n    company_id: str\n    owner: Owner\n\n# Now you can use the model to save it to Firestore\nowner = Owner(first_name="John", last_name="Doe")\ncompany = Company(company_id="1234567-8", owner=owner)\ncompany.save()\n\n# Prints out the firestore ID of the Company model\nprint(company.id)\n```\n\nQuerying is done via a MongoDB-like `find()`:\n\n```python\nfrom firedantic import Model\nimport firedantic.operators as op\n\nclass Product(Model):\n    __collection__ = "products"\n    product_id: str\n    stock: int\n\n\nProduct.find({"product_id": "abc-123"})\nProduct.find({"stock": {">=": 3}})\n# or\nProduct.find({"stock": {op.GTE: 3}})\n```\n\nThe query operators are found at [https://firebase.google.com/docs/firestore/query-data/queries#query_operators](https://firebase.google.com/docs/firestore/query-data/queries#query_operators).\n\n\n### Async usage\n\nFiredantic can also be used in an async way, like this:\n```python\nimport asyncio\nfrom os import environ\nfrom unittest.mock import Mock\n\nimport google.auth.credentials\nfrom google.cloud.firestore import AsyncClient\n\nfrom firedantic import AsyncModel, configure\n\n# Firestore emulator must be running if using locally.\nif environ.get("FIRESTORE_EMULATOR_HOST"):\n    client = AsyncClient(\n        project="firedantic-test",\n        credentials=Mock(spec=google.auth.credentials.Credentials),\n    )\nelse:\n    client = AsyncClient()\n\nconfigure(client, prefix="firedantic-test-")\n\n\nclass Person(AsyncModel):\n    __collection__ = "persons"\n    name: str\n\n\nasync def main():\n    alice = Person(name="Alice")\n    await alice.save()\n    print(f"Saved Alice as {alice.id}")\n    bob = Person(name="Bob")\n    await bob.save()\n    print(f"Saved Bob as {bob.id}")\n\n    found_alice = await Person.find_one({"name": "Alice"})\n    print(f"Found Alice: {found_alice.id}")\n    assert alice.id == found_alice.id\n\n    found_bob = await Person.get_by_id(bob.id)\n    assert bob.id == found_bob.id\n    print(f"Found Bob: {found_bob.id}")\n\n    await alice.delete()\n    print("Deleted Alice")\n    await bob.delete()\n    print("Deleted Bob")\n\n\nif __name__ == "__main__":\n    # Starting from Python 3.7 ->\n    # asyncio.run(main())\n\n    # Compatible with Python 3.6 ->\n    loop = asyncio.get_event_loop()\n    result = loop.run_until_complete(main())\n```\n\n## Subcollections\n\nSubcollections in Firestore are basically dynamically named collections.\n\nFiredantic supports them via the `SubCollection` and `SubModel` classes, by creating dynamic classes with collection name determined based on the "parent" class it is in reference to using the `model_for()` method.\n\n```python\nfrom typing import Optional, Type\n\nfrom firedantic import AsyncModel, AsyncSubCollection, AsyncSubModel, ModelNotFoundError\n\n\nclass UserStats(AsyncSubModel):\n    id: Optional[str]\n    purchases: int = 0\n\n    class Collection(AsyncSubCollection):\n        # Can use any properties of the "parent" model\n        __collection_tpl__ = "users/{id}/stats"\n\n\nclass User(AsyncModel):\n    __collection__ = "users"\n    name: str\n\n\nasync def get_user_purchases(user_id: str, period: str = "2021") -> int:\n    user = await User.get_by_id(user_id)\n    stats_model: Type[UserStats] = UserStats.model_for(user)\n    try:\n        stats = await stats_model.get_by_id(period)\n    except ModelNotFoundError:\n        stats = stats_model()\n    return stats.purchases\n\n```\n\n## Development\n\nPRs are welcome!\n\n\n\nTo run tests locally, you should run:\n\n```bash\npoetry install\npoetry run invoke test\n```\n\n### About sync and async versions of library\n\nAlthough this library provides both sync and async versions of models, please keep in mind that\nyou need to explicitly maintain only async version of it. The synchronous version is generated automatically by invoke task:\n\n```bash\npoetry run invoke unasync\n```\n\nWe decided to go this way in order to:\n- make sure both versions have the same API\n- reduce human error factor\n- avoid working on two code bases at the same time to reduce maintenance effort\n\nThus, please make sure you don\'t modify any of files under [firedantic/_sync](./firedantic/_sync) and [firedantic/tests/tests_sync](./firedantic/tests/tests_sync) by hands.\n`unasync` is also running as part of pre-commit hooks, but in order to run the latest version of tests you have to run it manually.\n\n## License\n\nThis code is released under the BSD 3-Clause license. Details in the\n[LICENSE](./LICENSE) file.\n',
    'author': 'Digital Living International Ltd',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/digitalliving/firedantic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



from uuid import uuid4

import pytest
from pydantic import Field, ValidationError

import firedantic.operators as op
from firedantic import Model
from firedantic.exceptions import (
    CollectionNotDefined,
    InvalidDocumentID,
    ModelNotFoundError,
)
from firedantic.tests.tests_sync.conftest import (
    Company,
    CustomIDConflictModel,
    CustomIDModel,
    CustomIDModelExtra,
    Product,
    TodoList,
    User,
    UserStats,
    get_user_purchases,
)

TEST_PRODUCTS = [
    {"product_id": "a", "stock": 0},
    {"product_id": "b", "stock": 1},
    {"product_id": "c", "stock": 2},
    {"product_id": "d", "stock": 3},
]


def test_save_model(configure_db, create_company):
    company = create_company()

    assert company.id is not None
    assert company.owner.first_name == "John"
    assert company.owner.last_name == "Doe"


def test_delete_model(configure_db, create_company):
    company: Company = create_company(
        company_id="11223344-5", first_name="Jane", last_name="Doe"
    )

    _id = company.id

    company.delete()

    with pytest.raises(ModelNotFoundError):
        Company.get_by_id(_id)


def test_find_one(configure_db, create_company):
    with pytest.raises(ModelNotFoundError):
        Company.find_one()

    company_a: Company = create_company(company_id="1234555-1", first_name="Foo")
    company_b: Company = create_company(company_id="1231231-2", first_name="Bar")

    a: Company = Company.find_one({"company_id": company_a.company_id})
    b: Company = Company.find_one({"company_id": company_b.company_id})

    assert a.company_id == company_a.company_id
    assert b.company_id == company_b.company_id
    assert a.owner.first_name == "Foo"
    assert b.owner.first_name == "Bar"

    with pytest.raises(ModelNotFoundError):
        Company.find_one({"company_id": "Foo"})

    random_company = Company.find_one()
    assert random_company.company_id in {a.company_id, b.company_id}


def test_find(configure_db, create_company, create_product):
    ids = ["1234555-1", "1234567-8", "2131232-4", "4124432-4"]
    for company_id in ids:
        create_company(company_id=company_id)

    c = Company.find({"company_id": "4124432-4"})
    assert c[0].company_id == "4124432-4"
    assert c[0].owner.first_name == "John"

    d = Company.find({"owner.first_name": "John"})
    assert len(d) == 4

    for p in TEST_PRODUCTS:
        create_product(**p)

    assert len(Product.find({})) == 4

    products = Product.find({"stock": {op.GTE: 1}})
    assert len(products) == 3

    products = Product.find({"stock": {op.GTE: 2, op.LT: 4}})
    assert len(products) == 2

    products = Product.find({"product_id": {op.IN: ["a", "d", "g"]}})
    assert len(products) == 2

    with pytest.raises(ValueError):
        Product.find({"product_id": {"<>": "a"}})


def test_find_not_in(configure_db, create_company):
    ids = ["1234555-1", "1234567-8", "2131232-4", "4124432-4"]
    for company_id in ids:
        create_company(company_id=company_id)

    found = Company.find(
        {
            "company_id": {
                op.NOT_IN: [
                    "1234555-1",
                    "1234567-8",
                ]
            }
        }
    )
    assert len(found) == 2
    for company in found:
        assert company.company_id in ("2131232-4", "4124432-4")


def test_find_array_contains(configure_db, create_todolist):
    list_1 = create_todolist("list_1", ["Work", "Eat", "Sleep"])
    create_todolist("list_2", ["Learn Python", "Walk the dog"])

    found = TodoList.find({"items": {op.ARRAY_CONTAINS: "Eat"}})
    assert len(found) == 1
    assert found[0].name == list_1.name


def test_find_array_contains_any(configure_db, create_todolist):
    list_1 = create_todolist("list_1", ["Work", "Eat"])
    list_2 = create_todolist("list_2", ["Relax", "Chill", "Sleep"])
    create_todolist("list_3", ["Learn Python", "Walk the dog"])

    found = TodoList.find({"items": {op.ARRAY_CONTAINS_ANY: ["Eat", "Sleep"]}})
    assert len(found) == 2
    for lst in found:
        assert lst.name in (list_1.name, list_2.name)


def test_get_by_id(configure_db, create_company):
    c: Company = create_company(company_id="1234567-8")

    assert c.id is not None
    assert c.company_id == "1234567-8"
    assert c.owner.last_name == "Doe"

    c_2 = Company.get_by_id(c.id)

    assert c_2.id == c.id
    assert c_2.company_id == "1234567-8"
    assert c_2.owner.first_name == "John"


def test_get_by_empty_str_id(configure_db):
    with pytest.raises(ModelNotFoundError):
        Company.get_by_id("")


def test_missing_collection(configure_db):
    class User(Model):
        name: str

    with pytest.raises(CollectionNotDefined):
        User(name="John").save()


def test_model_aliases(configure_db):
    class User(Model):
        __collection__ = "User"

        first_name: str = Field(..., alias="firstName")
        city: str

    user = User(firstName="John", city="Helsinki")
    user.save()

    user_from_db = User.get_by_id(user.id)
    assert user_from_db.first_name == "John"
    assert user_from_db.city == "Helsinki"


@pytest.mark.parametrize(
    "model_id",
    [
        "abc",
        pytest.param("a" * 1500, id="1500 chars"),
        "...",
        ".foo",
        "..bar",
        "f..oo",
        "bar..",
        "__",
        "___",
        "_foo_",
        "__bar_",
        "_baz__",
        "b__a__r",
        "å",
        "😀",
        " ",
        '"',
        "'",
        "\\",
        "\x00",
        "\x01",
        "\x07",
        "!:&+-*'()",
    ],
)
def test_models_with_valid_custom_id(configure_db, model_id):
    product_id = str(uuid4())

    product = Product(product_id=product_id, price=123.45, stock=2)
    product.id = model_id
    product.save()

    found = Product.get_by_id(model_id)
    assert found.product_id == product_id

    found.delete()


@pytest.mark.parametrize(
    "model_id",
    [
        "",
        pytest.param("a" * 1501, id="1501 chars"),
        ".",
        "..",
        "____",
        "__foo__",
        "__😀__",
        "/",
        "foo/bar",
        "foo/bar/baz",
    ],
)
def test_models_with_invalid_custom_id(configure_db, model_id):
    product = Product(product_id="product 123", price=123.45, stock=2)
    product.id = model_id
    with pytest.raises(InvalidDocumentID):
        product.save()

    with pytest.raises(ModelNotFoundError):
        Product.get_by_id(model_id)


def test_truncate_collection(configure_db, create_company):
    create_company(company_id="1234567-8")
    create_company(company_id="1234567-9")

    companies = Company.find({})
    assert len(companies) == 2

    Company.truncate_collection()
    new_companies = Company.find({})
    assert len(new_companies) == 0


def test_custom_id_model(configure_db):
    c = CustomIDModel(bar="bar")
    c.save()

    models = CustomIDModel.find({})
    assert len(models) == 1

    m = models[0]
    assert m.foo is not None
    assert m.bar == "bar"


def test_custom_id_conflict(configure_db):
    CustomIDConflictModel(foo="foo", bar="bar").save()

    models = CustomIDModel.find({})
    assert len(models) == 1

    m = models[0]
    assert m.foo != "foo"
    assert m.bar == "bar"


def test_model_id_persistency(configure_db):
    c = CustomIDConflictModel(foo="foo", bar="bar")
    c.save()

    c = CustomIDConflictModel.get_by_doc_id(c.id)
    c.save()

    assert len(CustomIDConflictModel.find({})) == 1


def test_bare_model_document_id_persistency(configure_db):
    c = CustomIDModel(bar="bar")
    c.save()

    c = CustomIDModel.get_by_doc_id(c.foo)
    c.save()

    assert len(CustomIDModel.find({})) == 1


def test_bare_model_get_by_empty_doc_id(configure_db):
    with pytest.raises(ModelNotFoundError):
        CustomIDModel.get_by_doc_id("")


def test_extra_fields(configure_db):
    CustomIDModelExtra(foo="foo", bar="bar", baz="baz").save()
    with pytest.raises(ValidationError):
        CustomIDModel.find({})


def test_company_stats(configure_db, create_company):
    company: Company = create_company(company_id="1234567-8")
    company_stats = company.stats()

    stats = company_stats.get_stats()
    stats.sales = 100
    stats.save()

    # Ensure the data can be still loaded
    loaded = company.stats().get_stats()
    assert loaded.sales == stats.sales

    # And that we can still save
    loaded.sales += 1
    loaded.save()

    stats = company_stats.get_stats()
    assert stats.sales == 101


def test_subcollection_model_safety(configure_db):
    """
    Ensure you shouldn't be able to use unprepared subcollection models accidentally
    """
    with pytest.raises(CollectionNotDefined):
        UserStats.find({})


def test_get_user_purchases(configure_db):
    u = User(name="Foo")
    u.save()

    us = UserStats.model_for(u)
    us(id="2021", purchases=42).save()

    assert get_user_purchases(u.id) == 42



import uuid
from typing import List, Optional, Type
from unittest.mock import Mock

import google.auth.credentials
import pytest
from google.cloud.firestore_v1 import AsyncClient
from pydantic import BaseModel, Extra, PrivateAttr

from firedantic import (
    AsyncBareModel,
    AsyncBareSubCollection,
    AsyncBareSubModel,
    AsyncModel,
    AsyncSubCollection,
    AsyncSubModel,
)
from firedantic.configurations import configure
from firedantic.exceptions import ModelNotFoundError


class CustomIDModel(AsyncBareModel):
    __collection__ = "custom"
    __document_id__ = "foo"

    foo: Optional[str]
    bar: str

    class Config:
        extra = Extra.forbid


class CustomIDModelExtra(AsyncBareModel):
    __collection__ = "custom"
    __document_id__ = "foo"

    foo: Optional[str]
    bar: str
    baz: str

    class Config:
        extra = Extra.forbid


class CustomIDConflictModel(AsyncModel):
    __collection__ = "custom"

    foo: str
    bar: str

    class Config:
        extra = Extra.forbid


class Owner(BaseModel):
    """Dummy owner Pydantic model."""

    first_name: str
    last_name: str

    class Config:
        extra = Extra.forbid


class CompanyStats(AsyncBareSubModel):
    _doc_id: Optional[str] = PrivateAttr()
    sales: int

    class Collection(AsyncBareSubCollection):
        __collection_tpl__ = "companies/{id}/Stats"
        __document_id__ = "_doc_id"

    @classmethod
    async def _get_by_id_or_empty(cls, _doc_id) -> AsyncBareSubModel:
        try:
            return await cls.get_by_doc_id(_doc_id)
        except ModelNotFoundError:
            model = cls._create(  # type: ignore
                sales=0,
            )
            model._doc_id = _doc_id
            return model  # type: ignore

    @classmethod
    async def get_stats(cls, period="2021"):
        return await cls._get_by_id_or_empty(period)


class Company(AsyncModel):
    """Dummy company Firedantic model."""

    __collection__ = "companies"
    company_id: str
    owner: Owner

    class Config:
        extra = Extra.forbid

    def stats(self) -> Type[CompanyStats]:
        return CompanyStats.model_for(self)  # type: ignore


class Product(AsyncModel):
    """Dummy product Firedantic model."""

    __collection__ = "products"
    product_id: str
    price: float
    stock: int

    class Config:
        extra = Extra.forbid


class TodoList(AsyncModel):
    __collection__ = "todoLists"
    name: str
    items: List[str]

    class Config:
        extra = Extra.forbid


@pytest.fixture
def configure_db():
    client = AsyncClient(
        project="firedantic-test",
        credentials=Mock(spec=google.auth.credentials.Credentials),
    )

    prefix = str(uuid.uuid4()) + "-"
    configure(client, prefix)


@pytest.fixture
def create_company():
    async def _create(
        company_id: str = "1234567-8", first_name: str = "John", last_name: str = "Doe"
    ):
        owner = Owner(first_name=first_name, last_name=last_name)
        company = Company(company_id=company_id, owner=owner)
        await company.save()
        return company

    return _create


@pytest.fixture
def create_product():
    async def _create(product_id: str = None, price: float = 1.23, stock: int = 3):
        if not product_id:
            product_id = str(uuid.uuid4())
        p = Product(product_id=product_id, price=price, stock=stock)
        await p.save()
        return p

    return _create


@pytest.fixture
def create_todolist():
    async def _create(name: str, items: List[str]):
        p = TodoList(name=name, items=items)
        await p.save()
        return p

    return _create


# Test case from README
class UserStats(AsyncSubModel):
    id: Optional[str]
    purchases: int = 0

    class Collection(AsyncSubCollection):
        # Can use any properties of the "parent" model
        __collection_tpl__ = "users/{id}/stats"


class User(AsyncModel):
    __collection__ = "users"
    name: str


async def get_user_purchases(user_id: str, period: str = "2021") -> int:
    user = await User.get_by_id(user_id)
    stats_model: Type[UserStats] = UserStats.model_for(user)
    try:
        stats = await stats_model.get_by_id(period)
    except ModelNotFoundError:
        stats = stats_model()
    return stats.purchases
